{-

24-bit Bitmap Loader (v4)
by Andre W B Furtado
awbf@cin.ufpe.br

This program is intended to be used with HOpenGL only.
This version is intended to be used with GHC-5.0X !

-}

module Awbfbmp where

import GL
import IO
import IOExts

------ Start of Bitmap Code -------

binAux :: String
binAux = "000000000000000000000000"

type AwbfBitmap = (GLsizei, GLsizei, PixelDescriptor)
type BmpList = [(GLubyte, GLubyte, GLubyte, GLubyte)]
type ColorList3 = [(GLubyte, GLubyte, GLubyte)]

--loads a bitmap from a file
loadBitmap :: FilePath -> Maybe ColorList3 -> IO AwbfBitmap
loadBitmap bmName invList = do
	bmFile <- openFileEx bmName (BinaryMode ReadMode)
	bmString <- hGetContents bmFile
	(bmW,bmH) <- getWH (myDrop 18 bmString)
	bmData <- getBmData (myDrop 54 bmString) (bmW,bmH) invList
	hClose bmFile             
	return (bmW,bmH,bmData)

 -- loads n bitmaps from n files	
loadBitmapList :: [(FilePath, Maybe ColorList3)] -> IO [AwbfBitmap]
loadBitmapList bmps = do
	bmList <- loadBmListAux bmps []
	return (reverse bmList)

loadBmListAux :: [(FilePath, Maybe ColorList3)] -> [AwbfBitmap] -> IO [AwbfBitmap]
loadBmListAux [] bmList = return (bmList)
loadBmListAux ((n,l):as) bmList = do
	bm <- loadBitmap n l
	loadBmListAux as (bm:bmList)

getWH :: String -> IO (GLsizei,GLsizei)
getWH (a:b:c:d:e:f:g:h:_) = do
	return ( (op (bin a) 0) + (op (bin b) 8) + (op (bin c) 16) + (op (bin d) 24),
	         (op (bin e) 0) + (op (bin f) 8) + (op (bin g) 16) + (op (bin h) 24))
 	         where bin x = toBinary(fromEnum x)
	               op x n = toDecimal(shiftLeft(binAux ++ (make0 (8 - (length x)) ++ x)) n)
getWH _ = error "Awbfbmp.getWH error: strange bitmap file"	              

getBmData :: String -> (GLsizei,GLsizei) -> Maybe ColorList3 -> IO PixelDescriptor
getBmData bmString (bmW,bmH) invList = 
	let colorList = makeColorList bmString (bmW,bmH) in
	newArray [Color4 r g b a | (r,g,b,a) <- addInvisiblity colorList invList] >>= \bmData ->
       	return (PixelDescriptor GL.Rgba UnsignedByte (castPtr bmData))
       	
addInvisiblity :: ColorList3 -> Maybe ColorList3 -> BmpList
addInvisiblity [] _ = []
addInvisiblity l Nothing = map (\(r,g,b) -> (r,g,b,255)) l
addInvisiblity ((r,g,b):as) i@(Just invList) | (r,g,b) `elem` invList = ((r,g,b,0):(addInvisiblity as i))
     			    		     | otherwise = ((r,g,b,255):(addInvisiblity as i))
     			    		     
makeColorList :: String -> (GLsizei,GLsizei) -> [(GLubyte, GLubyte, GLubyte)]
makeColorList bmString (bmW,bmH) = makeColorListAux (bmW `mod` 4) bmString (bmW*bmH) (bmW,bmW)
			
makeColorListAux :: GLsizei -> String -> GLsizei -> (GLsizei,GLsizei) -> [(GLubyte, GLubyte, GLubyte)]
makeColorListAux _ _ 0 _ = []
makeColorListAux x bmString totVert (0,bmW) = makeColorListAux x (myDrop x bmString) totVert (bmW,bmW)
makeColorListAux x (b:g:r:bmString) totVert (n,bmW) = (ord2 r,ord2 g,ord2 b): (makeColorListAux x bmString (totVert - 1) (n - 1,bmW))
makeColorListAux _ _ _ _ = error "Awbfbmp.makeColorListAux error: strange bitmap file"

shiftLeft :: String -> Int -> String
shiftLeft a 0 = a
shiftLeft (_:as) n = shiftLeft(as ++ "0") (n-1)
shiftLeft _ _ = []

toDecimal :: String -> GLsizei
toDecimal a = toDecimalAux (reverse a) 32

toDecimalAux :: String -> GLsizei -> GLsizei
toDecimalAux [] _ = 0
toDecimalAux _ 0 = 0
toDecimalAux (a:as) n
		| a == '0' = toDecimalAux as (n-1)
		| otherwise = pow2 (32 - n) + toDecimalAux as (n-1)
		
pow2 :: GLsizei -> GLsizei
pow2 0 = 1
pow2 n = 2 * pow2(n-1)

toBinary :: Int -> String
toBinary n
	| n < 2 = show n
	| otherwise = toBinary (n `div` 2) ++ (show (n `mod` 2))
	
make0 :: Int -> String
make0 0 = []
make0 n = '0':(make0 (n-1))

ord2 :: Char -> GLubyte
ord2 a = (toEnum.fromEnum) a

myDrop                :: GLsizei -> [a] -> [a]
myDrop 0 xs            = xs
myDrop _ []            = []
myDrop n (_:xs) | n>0  = myDrop (n-1) xs
myDrop _ _ = error "Awbfbmp.myDrop error: negative argument"